program inno7zip;
uses
  Windows,
  Classes,
  BigIni,
  ShellAPI,
  Registry,
  SysUtils;

function AddSlash(str: string) : string;
begin
  Result := str;
  if (str <> '') and (Str[length(str)] <> '\') then
    Result := str + '\';
end;

procedure ProcessMessages;
var
  MsgRec : TMsg;
begin
  while PeekMessage(MsgRec, 0 {FWindowHandle}, 0, 0, PM_REMOVE) do begin
     TranslateMessage(MsgRec);
     DispatchMessage(MsgRec)
  end;
end;

procedure Delay(msec:real);
var
   FirstTickCount:longint;
begin
   FirstTickCount:=GetTickCount;
   repeat
     sleep(100);
     ProcessMessages;
   until ((GetTickCount-FirstTickCount) >= int(msec));
end;

function FileExecuteWait(const FileName, Params, StartDir: string;
  InitialState: word): Integer;
var
  Info: TShellExecuteInfo;
  ExitCode: DWORD;
begin
  Result := -1;
  FillChar(Info, SizeOf(Info), 0);
  Info.cbSize := SizeOf(TShellExecuteInfo);
  with Info do begin
    fMask := SEE_MASK_NOCLOSEPROCESS;
    Wnd := FindWindow('Progman', nil);
    lpFile := PChar(FileName);
    lpParameters := PChar(Params);
    lpDirectory := PChar(StartDir);
    nShow := InitialState;
  end;
  if ShellExecuteEx(@Info) then begin
    repeat
      sleep(100);
      ProcessMessages;
      GetExitCodeProcess(Info.hProcess, ExitCode);
    until (ExitCode <> STILL_ACTIVE);
    Result := ExitCode;
  end
  else Result := -1;
end;

function FileCopy(sourceName, destName : string; overwrite : boolean) : boolean;
var
  fsource, fdest : TfileStream;
begin
 Result := False;
 if FileExists(destName) and (not overwrite) then exit;
 TRY
   fsource := TfileStream.create(sourceName, fmOpenRead or fmShareDenyWrite);
   fdest := TfileStream.create(destName, fmCreate);
   fdest.CopyFrom(fsource, fsource.size);
   Result := True;
 FINALLY
   fsource.free;
   fdest.free;
 END;
end;

function BuildSFX(sfxName, zipName, destName : string; overwrite : boolean) : boolean;
var
  fsfx, fzip, fdest : TfileStream;
  stub : string;
begin
 result := False;
 if FileExists(destName) and (not overwrite) then exit;
 TRY
   fsfx := TfileStream.create(sfxName, fmOpenRead or fmShareDenyWrite);
   fzip := TfileStream.create(zipName, fmOpenRead or fmShareDenyWrite);
   fdest := TfileStream.create(destName, fmCreate);
   fdest.CopyFrom(fsfx, fsfx.size);

   stub := ';!@Install@!UTF-8!' + #13#10 +
           'Title="Extract Files..."'+ #13#10 +
           'RunProgram="'+ExtractFileName(destName)+'"' + #13#10 +
           ';!@InstallEnd@!';
   fdest.writeBuffer(PChar(stub)^, length(stub));
   fdest.CopyFrom(fzip, fzip.size);
   result := true;
 FINALLY
   fsfx.free;
   fzip.free;
   fdest.free;
 END;
end;

var
   installFile, scriptFileName : string;
   scriptFile : TBigIniFile;
   sfx7zip, sfxParams, newFile, compLevel, delscript : string;
   reg : TRegistry;
   enabled : boolean;
begin
  scriptFileName := ParamStr(1);
  reg := TRegistry.Create;
  if reg.OpenKey('SOFTWARE\7-ZIP', false) then
    sfx7zip := AddSlash(reg.ReadString('Path'))
  else if reg.OpenKey('Software\Bjornar Henden\ISTool4\Prefs', false) then
    sfx7zip := AddSlash(reg.ReadString('7zDirectory'))
  else sfx7zip := '';
  reg.free;
  if FileExists(scriptFileName) then
  begin
    scriptFile := TBigIniFile.Create(scriptFileName);
    installFile := AddSlash(scriptFile.ReadString('Setup', 'OutputDir', '')) +
                   scriptFile.ReadString('Setup', 'OutputBaseFilename', '') + '.exe';

    compLevel := scriptFile.ReadString('Inno7zip', 'CompressionLevel', '9');
    sfxParams := scriptFile.ReadString('Inno7zip', '7zipParams', '-m0=BCJ2 -m1=LZMA:d23 -m2=LZMA:d19 -m3=LZMA:d19 -mb0:1 -mb0s1:2 -mb0s2:3 -mx='+compLevel);
    newFile   := scriptFile.ReadString('Inno7zip', 'CopyTo', '');
    delscript := lowerCase(trim(scriptFile.ReadString('Inno7zip', 'DeleteTmpScript', 'true')));
    enabled :=   lowerCase(trim(scriptFile.ReadString('Inno7zip', 'Enable7zip', 'true'))) <> 'false';
    scriptFile.free;

    if enabled and FileExists(scriptFileName) and FileExists(sfx7zip+'7za.exe') and FileExists(sfx7zip+'7zS.sfx') then
    begin
      FileExecuteWait(sfx7zip+'7za.exe', 'a -y '+sfxParams+' tmpsetup.7z '+ExtractFileName(installFile), ExtractFileDir(installFile), SW_SHOWNORMAL);
      BuildSFX(sfx7zip+'7zS.sfx', AddSlash(ExtractFileDir(installFile))+'tmpsetup.7z', installFile, true);
    end;
    if delscript='true' then
       deleteFile(scriptFileName);
    if newFile <> '' then
       FileCopy(installFile, newFile, true);
  end;
end.
